/*
 * $Id: hswitch.c,v 1.2 1998/10/15 12:31:01 adam Exp $
 */
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "tclrobot.h"

#define TAG_MAX_LEN 32

#define SPACECHR " \t\r\n\f"

#define DEBUG(x) 

static int skipSpace (const char *cp)
{
    int i = 0;
    while (strchr (SPACECHR, cp[i]))
        i++;
    return i;
}

static int skipTag (const char *cp, char *dst)
{
    int i;
    int j = 0;

    for (i=0; cp[i] && !strchr (SPACECHR "/>=", cp[i]); i++)
	if (j < TAG_MAX_LEN-1)
	{
	    dst[j] = tolower(cp[j]);
	    j++;
	}
    dst[j] = '\0';
    return i;
}

static int skipParm (const char *cp, char *name, char **value)
{
    int i = skipTag (cp, name);   
    *value = NULL;
    if (!i)
        return skipSpace (cp);
    i += skipSpace (cp + i);
    if (cp[i] == '=')
    {
        int v0, v1;
        i++;
        i += skipSpace (cp + i);
        if (cp[i] == '\"')
        {
            v0 = ++i;
            while (cp[i] != '\"' && cp[i])
                i++; 
            v1 = i;
            if (cp[i])
                i++;
        }
        else
        {
            v0 = i;
            while (cp[i] && !strchr (SPACECHR ">", cp[i]))
                i++;
            v1 = i;
        }
        *value = malloc (v1 - v0 + 1);
        memcpy (*value, cp + v0, v1-v0);
        (*value)[v1-v0] = '\0';
    }
    i += skipSpace (cp + i);
    return i;
}

struct tagParm {
    char name[TAG_MAX_LEN];
    char *value;
    struct tagParm *next;
};

struct tagInfo {
    int level;
    char *pattern;
    char *code;

    char name[TAG_MAX_LEN];
    const char *body_start;
    struct tagParm *tagParms;
};

static int tagLookup (struct tagInfo *tags, int tagNo, const char *tagString)
{
    int i;
    for (i = 0; i<tagNo; i++)
        if (!strcmp (tags[i].pattern, tagString))
             return i;
    return -1;
}

static int tagStart (struct tagInfo *tag, const char *tagString,
                     const char *cp)
{
    int i;
    char parm_name[TAG_MAX_LEN];
    char *parm_value;
    struct tagParm **nParms;

    if (tag && !tag->level)
    {
        strcpy (tag->name, tagString);
	DEBUG(printf ("------ consuming this %s\n", tag->name));
        tag->tagParms = NULL;
        nParms = &tag->tagParms;
    }

    i = skipSpace (cp);
    while (cp[i] && cp[i] != '>')
    {
        int nor = skipParm (cp+i, parm_name, &parm_value);
        i += nor;
	if (nor && tag)
	{
	    DEBUG(printf ("parm_name=%s parm_value=%s\n", parm_name, parm_value));
	}
        if (nor && tag && !tag->level)
        {
            *nParms = malloc (sizeof(**nParms));
            assert (*nParms);
            strcpy ((*nParms)->name, parm_name);
            (*nParms)->value = parm_value;
            (*nParms)->next = NULL;
	    nParms = &(*nParms)->next;
        }
        else
        {
            if (!nor)
                i++;
            free (parm_value);
        }
    }
    if (cp[i])
        i++;
    if (tag)
    {
        if (!tag->level)
            tag->body_start = cp+i;
        ++(tag->level);
    }
    return i;
}

static int tagEnd (Tcl_Interp *interp, struct tagInfo *tag,
                   const char *tagString, const char *cp, const char *body_end)
{
    int i = 0;

    if (cp[i] == '>')
        i++;

    if (tag && tag->level)
    {
        -- (tag->level);
        if (!tag->level)
        {
            struct tagParm *tp = tag->tagParms;
            char *value = malloc (body_end - tag->body_start + 1);

            assert (value);
            memcpy (value, tag->body_start, body_end - tag->body_start);
            value[body_end - tag->body_start] = '\0';
            Tcl_SetVar (interp, "body", value, 0);
            free (value);
            while (tp)
            {
                char vname[TAG_MAX_LEN+30];
                struct tagParm *tp0 = tp;
                
                sprintf (vname, "parm(%s)", tp->name);
		DEBUG(printf ("vname=%s\n", vname));

                Tcl_SetVar (interp, vname, tp->value ? tp->value : "",0);
                tp = tp->next;
                free (tp0);
            }
            Tcl_Eval (interp, tag->code);
        }
    }
    return i;
}

int htmlSwitch (ClientData clientData, Tcl_Interp *interp,
                int argc, char **argv)
{
    struct tagInfo *tags;
    int noTags;
    const char *cp;
    int i, argi = 1;

    cp = argv[argi++];
    noTags = (argc - argi)/2;
    if (noTags < 1)
    {
        interp->result =
            "wrong # args: should be ?switches? string pattern body ...";
        return TCL_ERROR;
    }
    tags = malloc (sizeof(*tags) * noTags);
    assert (tags);
    for (i = 0; i<noTags; i++)
    {
        tags[i].level = 0;
        tags[i].pattern = argv[argi++];
        tags[i].code = argv[argi++];
    }
    while (*cp)
    {
        if (cp[0] == '<' && cp[1] != '/')     /* start tag */
        {
            char tagStr[TAG_MAX_LEN];
            int tagI;

            cp++;
            cp += skipTag (cp, tagStr);
            tagI = tagLookup (tags, noTags, tagStr);
	    DEBUG(printf ("tagStr = %s tagI = %d\n", tagStr, tagI));
            cp += tagStart (tagI >= 0 ? tags+tagI : NULL, tagStr, cp);
        }
        else if (cp[0] == '<' && cp[1] == '/')/* end tag */
        {
            char tagStr[TAG_MAX_LEN];
            const char *body_end = cp;
            int tagI;

            cp += 2;
            cp += skipTag (cp, tagStr);
            tagI = tagLookup (tags, noTags, tagStr);
            cp += tagEnd (interp, tagI >= 0 ? tags+tagI : NULL,
                          tagStr, cp, body_end);
        }
        else                                  /* no tag */
            cp++;
    }
    free (tags);
    return TCL_OK;
}


